#include "WProgram.h"
void moistureCheck();
void wateringCheck();
void notify( uint8_t resp);
uint8_t serialavail_timeout(int timeout);
uint8_t readline_timeout(int timeout);
uint8_t XPort_reset(void);
uint8_t XPort_disconnected(void);
uint8_t XPort_connect(char *ipaddr, long port);
void XPort_flush(int timeout);
uint8_t posttweet(char *tweet);
void base64encode(char *s, char *r);
#include "AFSoftSerial.h"
#include "avr/io.h"
#include "string.h"
// Botanicalls helps plants ask for help.
// Rob Faludi with additional code from various public examples
// and from LadyAda's Twitter and Software Serial examples
// http://www.botanicalls.com
// Botanicalls is a project with Kati London, Rob Faludi, Kate Hartman and Rebecca Bray

char vers[] = "2.00"; // this version uses an XPort Shield directly


///// BOTANICALLS DEFINTIONS ////////
int wateredInterval = 30; // watering event interval in seconds
unsigned int moistInterval = 10;//moisture interval in seconds


#define MOIST 450 // minimum level of satisfactory moisture
#define DRY 350  // maximum level of tolerable dryness
#define SOAKED 600 // minimum desired level after watering
#define WATERING_CRITERIA 100 // minimum change in value that indicates watering

#define MOIST_SAMPLES 10 //number of moisture samples to average

int moistValues[MOIST_SAMPLES];

#define LEDPIN 13 // generic status LED
#define MOISTPIN 0 // moisture input is on analog pin 0
#define MOISTLED 9  // LED that indicates the plant needs water

#define MOIST_SAMPLE_INTERVAL 10 // seconds over which to average moisture samples
#define WATERED_INTERVAL 20 // seconds between checks for watering events

#define TWITTER_INTERVAL 1// minimum seconds between twitter postings

unsigned long lastMoistTime=0;
unsigned long lastWaterTime=0;
unsigned long lastTwitterTime=0;

int lastMoistAvg=1023;
int lastWaterVal=1023;


///// TWITTER DEFINITIONS ///////
#define IPADDR "128.121.146.100"  // twitter.com
#define PORT 80                   // HTTP
#define USERNAMEPASS "faludi:interact"  // your twitter username and password, seperated by a :
#define HTTPPATH "/statuses/update.xml"      // the person we want to follow

#define TWEETLEN 141
char linebuffer[256]; // oi
int lines = 0;

#define XPORT_RXPIN 3
#define XPORT_TXPIN 2
#define XPORT_RESETPIN 4
#define XPORT_DTRPIN 5
#define XPORT_CTSPIN 6
#define XPORT_RTSPIN 7

#define ERROR_NONE 0

#define ERROR_TIMEDOUT 2
#define ERROR_BADRESP 3
#define ERROR_DISCONN 4
uint8_t errno;

AFSoftSerial mySerial =  AFSoftSerial(XPORT_RXPIN, XPORT_TXPIN);

uint32_t laststatus = 0, currstatus = 0;

/************************SKETCH START*********************/

void setup()  { 
  

  
  uint8_t ret;
  
  pinMode(LEDPIN, OUTPUT);
  pinMode(MOISTLED, OUTPUT);
  
    for(int i = 0; i < MOIST_SAMPLES; i++) {
    moistValues[i] = 0; 
  }

  Serial.begin(9600);
  // set the data rate for the SoftwareSerial port
  mySerial.begin(9600);
  Serial.println("");
  Serial.println("Botanicalls starting...");

  // xport
  pinMode(XPORT_RESETPIN, OUTPUT);
  if (XPORT_DTRPIN) {
    pinMode(XPORT_DTRPIN, INPUT);
  }
  if (XPORT_CTSPIN) {
    pinMode(XPORT_CTSPIN, OUTPUT);
  }
  if (XPORT_RTSPIN) {
    pinMode(XPORT_RTSPIN, INPUT);
  }
  
 uint8_t response = posttweet("Botanicalls!!");  // tweet tweet
  if (response) 
     Serial.println("twot successful");
}


void loop()       // main loop of the program     
{

moistureCheck();
wateringCheck(); 
  
  
}




//function for checking soil moisture against threshold
void moistureCheck() {
  static byte returnValue = 0;
  static int counter = 1;//init static counter
  int moistAverage = 0; // init soil moisture average
  if((millis() - lastMoistTime) / 1000 > (MOIST_SAMPLE_INTERVAL / MOIST_SAMPLES)) {
    for(int i = MOIST_SAMPLES - 1; i > 0; i--) {
      moistValues[i] = moistValues[i-1]; //move the first measurement to be the second one, and so forth until we reach the end of the array.   
    }
    moistValues[0] = analogRead(MOISTPIN);//take a measurement and put it in the first place
    lastMoistTime = millis();
    int moistTotal = 0;//create a little local int for an average of the moistValues array
    for(int i = 0; i < MOIST_SAMPLES; i++) {//average the measurements (but not the nulls)
      moistTotal += moistValues[i];//in order to make the average we need to add them first 
    }
    if(counter<MOIST_SAMPLES) {
      moistAverage = moistTotal/counter;
      counter++; //this will add to the counter each time we've gone through the function
    }
    else {
      moistAverage = moistTotal/MOIST_SAMPLES;//here we are taking the total of the current light readings and finding the average by dividing by the array size
    } 
    //lastMeasure = millis();
    Serial.print("moisture: ");
    Serial.println(moistAverage,DEC); 

    ///return values
    if ((moistAverage < DRY)  &&  (lastMoistAvg >= DRY)  &&  (millis() > (lastTwitterTime + TWITTER_INTERVAL)) ) {
      uint8_t response = posttweet("URGENT! Water me!");  // tweet tweet
      notify(response); 
    }
    else if  ((moistAverage < MOIST)  &&  (lastMoistAvg >= MOIST)  &&  (millis() > (lastTwitterTime + TWITTER_INTERVAL)) ) {
      uint8_t response = posttweet("Water me please");  // tweet tweet
      notify(response); 
    }
    lastMoistAvg = moistAverage;
  }
}


//function for checking for watering events
void wateringCheck() {
  static byte returnValue = 0;
  int moistAverage = 0; // init soil moisture average
  if((millis() - lastWaterTime) / 1000 > WATERED_INTERVAL) {

    int waterVal = analogRead(MOISTPIN);//take a moisture measurement
    lastWaterTime = millis();

    Serial.print("watered: ");
    Serial.println(waterVal,DEC);
    if (waterVal >= lastWaterVal + WATERING_CRITERIA) { // if we've detected a watering event
      if (waterVal >= SOAKED  &&  lastWaterVal < MOIST &&  (millis() > (lastTwitterTime + TWITTER_INTERVAL))) {
        uint8_t response = posttweet("Thank you for watering me");  // tweet tweet
        notify(response); 
      }
      else if  (waterVal >= SOAKED  &&  lastWaterVal >= MOIST  &&  (millis() > (lastTwitterTime + TWITTER_INTERVAL)) ) {
        uint8_t response = posttweet("You over watered me");  // tweet tweet
        notify(response); 
      }
      else if  (waterVal < SOAKED  &&  lastWaterVal < MOIST  &&  (millis() > (lastTwitterTime + TWITTER_INTERVAL)) ) {
        uint8_t response = posttweet("You didn't water me enough");  // tweet tweet
        notify(response); 
      }
    }    
    lastWaterVal = waterVal;
  }
}


void notify( uint8_t resp) {
  if (resp)
  Serial.println("twitter sent");
  else {
    Serial.println("twitter failed");
  }
}

/***********************SOFTWARE UART*************************/

uint8_t serialavail_timeout(int timeout) {  // in ms
 
  while (timeout) {
    if (mySerial.available()) {
      if (XPORT_CTSPIN) { // we read some stuff, time to stop!
        digitalWrite(XPORT_CTSPIN, HIGH);
      }
      return 1;
    }
    // nothing in the queue, tell it to send something
    if (XPORT_CTSPIN) {
     digitalWrite(XPORT_CTSPIN, LOW);
    }
    timeout -= 1;
    delay(1);
  }
  if (XPORT_CTSPIN) { // we may need to process some stuff, so stop now
     digitalWrite(XPORT_CTSPIN, HIGH);
  }
  return 0;
}

uint8_t readline_timeout(int timeout) {
  uint8_t idx=0;
  char c;
  while (serialavail_timeout(timeout)) {
      c = mySerial.read();
      linebuffer[idx++] = c;
      if ((c == '\n') || (idx == 255)) {
        linebuffer[idx] = 0;
        errno = ERROR_NONE;
        return idx;
      }
  }
  linebuffer[idx] = 0;
  errno = ERROR_TIMEDOUT;
  return idx;
}

/********************XPORT STUFF**********************/

uint8_t XPort_reset(void) {
  char d;
  
  // 200 ms reset pulse
  digitalWrite(XPORT_RESETPIN, LOW);
  delay(200);
  digitalWrite(XPORT_RESETPIN, HIGH);
  
  // wait for 'D' for disconnected
  if (serialavail_timeout(5000)) { // 3 second timeout 
    d = mySerial.read();
    //Serial.print("Read: "); Serial.print(d, HEX);
    if (d != 'D'){
      return ERROR_BADRESP;
    } else {
      return 0;
    }
  }
  return ERROR_TIMEDOUT;
}  

uint8_t XPort_disconnected(void) {
  if (XPORT_DTRPIN != 0) {
     return digitalRead(XPORT_DTRPIN);
  } 
  return 0;
}


uint8_t XPort_connect(char *ipaddr, long port) {
  char ret;
  
  mySerial.print('C');
  mySerial.print(ipaddr);
  mySerial.print('/');
  mySerial.println(port);
  // wait for 'C'
  if (serialavail_timeout(3000)) { // 3 second timeout 
    ret = mySerial.read();
    Serial.print("Read: "); Serial.print(ret, HEX);
    if (ret != 'C') {
      return ERROR_BADRESP;
    }
  } else { 
    return ERROR_TIMEDOUT; 
  }
  return 0;
}

void XPort_flush(int timeout) {
  while (serialavail_timeout(timeout)) {
      mySerial.read();
  }
}

/********************TWITTER STUFF**********************/

uint8_t posttweet(char *tweet) {
  uint8_t ret;
  uint8_t success = 0;

  ret = XPort_reset();
  //Serial.print("Ret: "); Serial.print(ret, HEX);
  switch (ret) {
   case  ERROR_TIMEDOUT: { 
      Serial.println("Timed out on reset!"); 
      return 0;
   }
   case ERROR_BADRESP:  { 
      Serial.println("Bad response on reset!");
      return 0;
   }
   case ERROR_NONE: { 
    Serial.println("Reset OK!");
    break;
   }
   default:
     Serial.println("Unknown error"); 
     return 0;
  }
  
  // time to connect...
 
  ret = XPort_connect(IPADDR, PORT);
    switch (ret) {
   case  ERROR_TIMEDOUT: { 
      Serial.println("Timed out on connect"); 
      return 0;
   }
   case ERROR_BADRESP:  { 
      Serial.println("Failed to connect");
      return 0;
   }
   case ERROR_NONE: { 
     Serial.println("Connected..."); break;
   }
   default:
     Serial.println("Unknown error"); 
     return 0;
  }

  base64encode(USERNAMEPASS, linebuffer);
  
  // send the HTTP command, ie "GET /username/"
  mySerial.print("POST "); mySerial.print(HTTPPATH); mySerial.println(" HTTP/1.1");
  Serial.print("POST "); Serial.print(HTTPPATH); Serial.println(" HTTP/1.1");
  // next, the authentication
  mySerial.print("Host: "); mySerial.println(IPADDR);
  Serial.print("Host: "); Serial.println(IPADDR);
  mySerial.print("Authorization: Basic ");
  Serial.print("Authorization: Basic ");
  mySerial.println(linebuffer);
  Serial.println(linebuffer);
  mySerial.print("Content-Length: "); mySerial.println(7+strlen(tweet), DEC);
  Serial.print("Content-Length: "); Serial.println(7+strlen(tweet), DEC);
  mySerial.print("\nstatus="); mySerial.println(tweet);
  Serial.print("\nstatus="); Serial.println(tweet);
  
  mySerial.println("");  
 
  while (1) {
    // read one line from the xport at a time
    ret = readline_timeout(3000); // 3s timeout
    // if we're using flow control, we can actually dump the line at the same time!
    Serial.print(linebuffer);
    if (strstr(linebuffer, "HTTP/1.1 200 OK") == linebuffer)
       success = 1;
    
    if (((errno == ERROR_TIMEDOUT) && XPort_disconnected()) ||
        ((XPORT_DTRPIN == 0) &&
         (linebuffer[0] == 'D') && (linebuffer[1] == 0)))  {
       Serial.println("\nDisconnected...");
       return success;
    }
  }
}


void base64encode(char *s, char *r) {
  char padstr[4];
  char base64chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  uint8_t i, c;
  uint32_t n;
  
  c = strlen(s) % 3;
  if (c > 0) { 
    for (i=0; c < 3; c++) { 
      padstr[i++] = '='; 
    } 
  }
  padstr[i]=0;

  i = 0;
  for (c=0; c < strlen(s); c+=3) { 
     // these three 8-bit (ASCII) characters become one 24-bit number
     n = s[c]; n <<= 8;
     n += s[c+1]; 
     if (c+2 > strlen(s)) {
       n &= 0xff00;
     }
     n <<= 8;
     n += s[c+2];
     if (c+1 > strlen(s)) {
       n &= 0xffff00;
     }
       
     // this 24-bit number gets separated into four 6-bit numbers
     // those four 6-bit numbers are used as indices into the base64 character list
     r[i++] = base64chars[(n >> 18) & 63];
     r[i++] = base64chars[(n >> 12) & 63];
     r[i++] = base64chars[(n >> 6) & 63];
     r[i++] = base64chars[n & 63];
  }
  i -= strlen(padstr);
  for (c=0; c<strlen(padstr); c++) {
    r[i++] = padstr[c];  
  }
  r[i] = 0;
  Serial.println(r);
}


int main(void)
{
	init();

	setup();
    
	for (;;)
		loop();
        
	return 0;
}

